/**
 **	Imagewin.cc - A window to blit images into.
 **
 **	Written: 8/13/98 - JSF
 **/

/*
Copyright (C) 1998 Jeffrey S. Freedman

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.
*/

#include "imagewin.h"
#include <string.h>
#include <X11/Xutil.h>
#include <iostream.h>

/*
 *	Create a window with the proper depth.
 *
 *	Output:	->proper Image_window, or 0 if depth not supported.
 */

Image_window *Image_window::create
	(
	Display *dpy,
	int screen_num,
	unsigned int w,			// Desired width, height.
	unsigned int h,
	unsigned long bdr_color,	// Colors for border, background.
	unsigned long back_color
	)
	{
					// Create image to work with.
	int depth = DisplayPlanes(dpy, screen_num);
	switch (depth)			// What depth?
		{
	case 8:	
		return (new Image_window8(dpy, screen_num, w, h,
				bdr_color, back_color));
	case 15:
	case 16:
		return (new Image_window16(dpy, screen_num, w, h,
				bdr_color, back_color));
	case 32:
		return (0);		// Later.
		}
	}

/*
 *	Create window.
 */

Image_window::Image_window
	(
	Display *dpy,
	int screen_num,
	unsigned int w,			// Desired width, height.
	unsigned int h,
	unsigned long bdr_color,	// Colors for border, background.
	unsigned long back_color
	) : display(dpy), width(w), height(h), clipx(0), clipy(0),
	    clipw(w), cliph(h), image(0), bits(0), line_width(w)
	{
					// Get color display type.
	visual = DefaultVisual(display, screen_num);
					// Create image to work with.
	depth = DisplayPlanes(display, screen_num);
	switch (depth)			// What depth?
		{
	case 8:	
		pixel_size = 1; break;
	case 15:
	case 16:
		pixel_size = 2; break;
	case 32:
		pixel_size = 4; break;
		}
					// Shared memory available?
	int shm_major,shm_minor, shm_pixmaps, shm_completion;
	shm = XShmQueryVersion (display, &shm_major, &shm_minor, &shm_pixmaps);
	XSetWindowAttributes atts;	// Set up attributes.
	atts.border_pixel = bdr_color;
	atts.background_pixel = back_color;
	win = XCreateWindow(display, 
		RootWindow(display, screen_num), 0, 0, width, height,
		4,			// Border width.
		DisplayPlanes(display, screen_num),
		CopyFromParent, CopyFromParent, CWBackPixel | CWBorderPixel,
		&atts);
					// Get graphics context.
	XGCValues gcvalues;
	gc = XCreateGC(display, win, 0L, &gcvalues);
					// Set foreground color for drawing.
	XSetForeground(display, gc, BlackPixel(display, screen_num));
	}

/*
 *	Destroy window.
 */

Image_window::~Image_window
	(
	)
	{
	XFreeGC(display, gc);
	free_image();
	}

/*
 *	Free the image.
 */

void Image_window::free_image
	(
	)
	{
	if (!image)
		return;
	if (shm)			// MIT shared mem?
		{			// Detach memory.
		XShmDetach(display, &shm_info);
					// Mark for deletion.
		shmctl(shm_info.shmid, IPC_RMID, 0);
		shmdt(shm_info.shmaddr);
		}
	else
		delete bits;
	bits = image->data = 0;
	XDestroyImage(image);
	image = 0;
	}

/*
 *	Window was resized.
 */

void Image_window::resized
	(
	Display *dpy,
	int screen_num,
	unsigned int neww, 
	unsigned int newh
	)
	{
	if (image)
		{
		if (neww == width && newh == height)
			return;		// Nothing changed.
		free_image();		// Delete old image.
		}
	width = neww;
	height = newh;
	if (shm)
		{
		image = XShmCreateImage(display, 0, 
			depth, ZPixmap, 0,
			&shm_info, width, height);
					// Get shared memory.
		shm_info.shmid = shmget(IPC_PRIVATE,
                                image->bytes_per_line * image->height,
                                IPC_CREAT|0777);
        	if (shm_info.shmid < 0 ||
		    (shm_info.shmaddr = (char *) shmat(shm_info.shmid, 0, 0))
							== ((char *) -1))
			shm = 0;
		else
			{
			bits = image->data = shm_info.shmaddr;
			shm_info.readOnly = False;
			XShmAttach(display, &shm_info);
			cout << "Using MIT shared memory ext.\n";
			}
		}
	if (!shm)
		{			// Create new image.
		int line_size = width * pixel_size;
		if (line_size & 3)	// Round up to nearest 32-bits.
			line_size += (4 - (line_size & 3));
		bits = new char[line_size * height];
		image = XCreateImage(display, 
			visual,
			depth,
			ZPixmap, 0, bits, width, height,
			32, line_size);
		}
					// Update line size in words.
	line_width = image->bytes_per_line/pixel_size;
	cout << "Bytes_per_line = " << image->bytes_per_line <<
			" Width = " << width << '\n';
	}

/*
 *	Repaint window.
 */

void Image_window::show
	(
	)
	{
	if (!ready())
		return;
	if (shm)
		XShmPutImage(display, win, gc, image, 0, 0, 0, 0,
				width, height, True);
	else
		XPutImage(display, win, gc, image, 0, 0, 0, 0, width, height);
	}

/*
 *	Exit if 16-bit color methods not supported.
 */

void Image_window::exit16
	(
	)
	{
	cerr << "16-bit color methods are not supported on this display.\n";
	free_image();
	exit(16);
	}

/*
 *	Create a default palette for 8-bit graphics on a 16-bit display.
 */

void Image_window16::create_default_palette
	(
	)
	{
	delete palette;			// Delete old.
	palette = new unsigned short[256];
	for (int i = 0; i < 256; i++)
		palette[i] = rgb(4*((i >> 5) & 0x7), 
			4*((i >> 2) & 0x7), 9*(i & 0x3));
	}

/*
 *	Fill with a given 16-bit value.
 */

void Image_window16::fill16
	(
	unsigned short pix
	)
	{
	unsigned short *pixels = get_pixels();
	int cnt = line_width*height;
	for (int i = 0; i < cnt; i++)
		*pixels++ = pix;
	}

/*
 *	Fill a rectangle with a 16-bit value.
 */

void Image_window16::fill16
	(
	unsigned short pix,
	int srcw, int srch,
	int destx, int desty
	)
	{
	int srcx = 0, srcy = 0;
					// Constrain to window's space.
	if (!clip(srcx, srcy, srcw, srch, destx, desty))
		return;
	unsigned short *pixels = get_pixels() + desty*line_width + destx;
	int to_next = line_width - srcw;// # pixels to next line.
	while (srch--)			// Do each line.
		{
		for (int cnt = srcw; cnt; cnt--)
			*pixels++ = pix;
		pixels += to_next;	// Get to start of next line.
		}
	}

/*
 *	Copy another rectangle into this one.
 */

void Image_window16::copy16
	(
	unsigned short *src_pixels,	// Source rectangle pixels.
	int srcw, int srch,		// Dimensions of source.
	int destx, int desty
	)
	{
	int srcx = 0, srcy = 0;
	int src_width = srcw;		// Save full source width.
					// Constrain to window's space.
	if (!clip(srcx, srcy, srcw, srch, destx, desty))
		return;
	unsigned short *to = get_pixels() + desty*line_width + destx;
	unsigned short *from = src_pixels + srcy*src_width + srcx;
	int to_next = line_width - srcw;// # pixels to next line.
	int from_next = src_width - srcw;
	while (srch--)			// Do each line.
		{
		for (int cnt = srcw; cnt; cnt--)
			*to++ = *from++;
		to += to_next;
		from += from_next;
		}
	}

/*
 *	Copy an area of the image within itself.
 */

void Image_window16::copy
	(
	int srcx, int srcy,		// Where to start.
	int srcw, int srch,		// Dimensions to copy.
	int destx, int desty		// Where to copy to.
	)
	{
	int ynext, yfrom, yto;		// Figure y stuff.
	if (srcy >= desty)		// Moving up?
		{
		ynext = line_width;
		yfrom = srcy;
		yto = desty;
		}
	else				// Moving down.
		{
		ynext = -line_width;
		yfrom = srcy + srch - 1;
		yto = desty + srch - 1;
		}
	unsigned short *to = get_pixels() + yto*line_width + destx;
	unsigned short *from = get_pixels() + yfrom*line_width + srcx;
					// Go through lines.
	while (srch--)
		{
		memmove((char *) to, (char *) from, srcw * 2);
		to += ynext;
		from += ynext;
		}
	}

/*
 *	Convert rgb value.
 */

inline unsigned char Get_color16
	(
	unsigned char val,
	int maxval,
	int brightness			// 100=normal.
	)
	{
	unsigned int c = (((unsigned int) val)*brightness*32)/
							(100*(maxval + 1));
	return (c < 32 ? c : 31);
	}

/*
 *	Set palette.
 */

void Image_window16::set_palette
	(
	unsigned char *rgbs,		// 256 3-byte entries.
	int maxval,			// Highest val. for each color.
	int brightness			// Brightness control (100 = normal).
	)
	{
					// Get the colors.
	for (int i = 0; i < 3*256; i += 3)
		{
		unsigned char r = Get_color16(rgbs[i], maxval, brightness);
		unsigned char g = Get_color16(rgbs[i + 1], maxval, brightness);
		unsigned char b = Get_color16(rgbs[i + 2], maxval, brightness);
		set_palette_color(i/3, r, g, b);
		}
	}



/*
 *	Copy another rectangle into this one.
 */

void Image_window16::copy8
	(
	unsigned char *src_pixels,	// Source rectangle pixels.
	int srcw, int srch,		// Dimensions of source.
	int destx, int desty
	)
	{
	int srcx = 0, srcy = 0;
	int src_width = srcw;		// Save full source width.
					// Constrain to window's space.
	if (!clip(srcx, srcy, srcw, srch, destx, desty))
		return;
	unsigned short *to = get_pixels() + desty*line_width + destx;
	unsigned char *from = src_pixels + srcy*src_width + srcx;
	int to_next = line_width - srcw;// # pixels to next line.
	int from_next = src_width - srcw;
	while (srch--)			// Do each line.
		{
		for (int cnt = srcw; cnt; cnt--)
			*to++ = palette[*from++];
		to += to_next;
		from += from_next;
		}
	}

/*
 *	Copy another rectangle into this one, with 0 being the transparent
 *	color.
 */

void Image_window16::copy_transparent8
	(
	unsigned char *src_pixels,	// Source rectangle pixels.
	int srcw, int srch,		// Dimensions of source.
	int destx, int desty
	)
	{
	int srcx = 0, srcy = 0;
	int src_width = srcw;		// Save full source width.
					// Constrain to window's space.
	if (!clip(srcx, srcy, srcw, srch, destx, desty))
		return;
	unsigned short *to = get_pixels() + desty*line_width + destx;
	unsigned char *from = src_pixels + srcy*src_width + srcx;
	int to_next = line_width - srcw;// # pixels to next line.
	int from_next = src_width - srcw;
	while (srch--)			// Do each line.
		{
		for (int cnt = srcw; cnt; cnt--, to++)
			{
			register int chr = *from++;
			if (chr)
				*to = palette[chr];
			}
		to += to_next;
		from += from_next;
		}
	}

/*
 *	Copy an area of the image within itself.
 */

void Image_window8::copy
	(
	int srcx, int srcy,		// Where to start.
	int srcw, int srch,		// Dimensions to copy.
	int destx, int desty		// Where to copy to.
	)
	{
	int ynext, yfrom, yto;		// Figure y stuff.
	if (srcy >= desty)		// Moving up?
		{
		ynext = line_width;
		yfrom = srcy;
		yto = desty;
		}
	else				// Moving down.
		{
		ynext = -line_width;
		yfrom = srcy + srch - 1;
		yto = desty + srch - 1;
		}
	char *to = bits + yto*line_width + destx;
	char *from = bits + yfrom*line_width + srcx;
					// Go through lines.
	while (srch--)
		{
		memmove((char *) to, (char *) from, srcw);
		to += ynext;
		from += ynext;
		}
	}

/*
 *	Convert rgb value.
 */

inline unsigned short Get_color8
	(
	unsigned char val,
	int maxval,
	int brightness			// 100=normal.
	)
	{
	unsigned long c = (((unsigned long) val)*brightness*65535L)/
							(100*(maxval + 1));
	return (c <= 65535L ? (unsigned short) c : 65535);
	}

/*
 *	Set palette.
 */

void Image_window8::set_palette
	(
	unsigned char *rgbs,		// 256 3-byte entries.
	int maxval,			// Highest val. for each color.
	int brightness			// Brightness control (100 = normal).
	)
	{
	XColor colors[256];
					// Get the colors.
	for (int i = 0; i < 256; i++)
		{
		colors[i].pixel = i;
		colors[i].flags = DoRed | DoGreen | DoBlue;
		colors[i].red = Get_color8(rgbs[3*i], maxval, brightness);
		colors[i].green = Get_color8(rgbs[3*i + 1], maxval, 
							brightness);
		colors[i].blue = Get_color8(rgbs[3*i + 2], maxval,
							brightness);
		}
	if (!colormap)			// First time?
		{
		colormap = XCreateColormap(display, win, visual, AllocAll);
		XSetWindowColormap(display, win, colormap);
		}
	XStoreColors(display, colormap, colors, 256);
	}

/*
 *	Fill with a given 8-bit value.
 */

void Image_window8::fill8
	(
	unsigned char pix
	)
	{
	char *pixels = bits;
	int cnt = line_width*height;
	for (int i = 0; i < cnt; i++)
		*pixels++ = pix;
	}

/*
 *	Fill a rectangle with an 8-bit value.
 */

void Image_window8::fill8
	(
	unsigned char pix,
	int srcw, int srch,
	int destx, int desty
	)
	{
	int srcx = 0, srcy = 0;
					// Constrain to window's space.
	if (!clip(srcx, srcy, srcw, srch, destx, desty))
		return;
	unsigned char *pixels = (unsigned char *) bits + 
						desty*line_width + destx;
	int to_next = line_width - srcw;// # pixels to next line.
	while (srch--)			// Do each line.
		{
		for (int cnt = srcw; cnt; cnt--)
			*pixels++ = pix;
		pixels += to_next;	// Get to start of next line.
		}
	}

/*
 *	Copy another rectangle into this one.
 */

void Image_window8::copy8
	(
	unsigned char *src_pixels,	// Source rectangle pixels.
	int srcw, int srch,		// Dimensions of source.
	int destx, int desty
	)
	{
	int srcx = 0, srcy = 0;
	int src_width = srcw;		// Save full source width.
					// Constrain to window's space.
	if (!clip(srcx, srcy, srcw, srch, destx, desty))
		return;
	char *to = bits + desty*line_width + destx;
	unsigned char *from = src_pixels + srcy*src_width + srcx;
	int to_next = line_width - srcw;// # pixels to next line.
	int from_next = src_width - srcw;
	while (srch--)			// Do each line.
		{
		for (int cnt = srcw; cnt; cnt--)
			*to++ = *from++;
		to += to_next;
		from += from_next;
		}
	}

/*
 *	Copy another rectangle into this one, with 0 being the transparent
 *	color.
 */

void Image_window8::copy_transparent8
	(
	unsigned char *src_pixels,	// Source rectangle pixels.
	int srcw, int srch,		// Dimensions of source.
	int destx, int desty
	)
	{
	int srcx = 0, srcy = 0;
	int src_width = srcw;		// Save full source width.
					// Constrain to window's space.
	if (!clip(srcx, srcy, srcw, srch, destx, desty))
		return;
	char *to = bits + desty*line_width + destx;
	unsigned char *from = src_pixels + srcy*src_width + srcx;
	int to_next = line_width - srcw;// # pixels to next line.
	int from_next = src_width - srcw;
	while (srch--)			// Do each line.
		{
		for (int cnt = srcw; cnt; cnt--, to++)
			{
			register int chr = *from++;
			if (chr)
				*to = chr;
			}
		to += to_next;
		from += from_next;
		}
	}

