/**
 **	Imagewin.h - A window to blit images into.
 **
 **	Written: 8/13/98 - JSF
 **/

/*
Copyright (C) 1998 Jeffrey S. Freedman

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.
*/

#ifndef INCL_IMAGEWIN
#define INCL_IMAGEWIN	1

#include <X11/Xlib.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

class Font_face;

/*
 *	Here's a window for showing and manipulating images.
 */
class Image_window
	{
	int shm;			// 1 if using MITSHM.
	XShmSegmentInfo shm_info;	// Shared memory info.
protected:
	Display *display;		// X-windows display.
	Window win;			// The window handle.
	GC gc;				// Graphics context for drawing.
	Visual *visual;			// Color characteristics.
	/*
	 *	Image info.
	 */
	unsigned int width, height;	// Dimensions (in pixels).
	int depth;			// # bits/pixel.
	int pixel_size;			// # bytes/pixel.
	char *bits;			// Allocated image buffer.
	XImage *image;			// Image containing 'bits'.
	unsigned int line_width;	// # words/scan-line.

	int clipx, clipy, clipw, cliph;	// Clip rectangle.
					// Clip.  Rets. 0 if nothing to draw.
	int clip(int& srcx, int& srcy, int& srcw, int& srch, 
							int& destx, int& desty)
		{
		if (destx < clipx)
			{
			if ((srcw += (destx - clipx)) <= 0)
				return (0);
			srcx -= (destx - clipx);
			destx = clipx;
			}
		if (desty < clipy)
			{
			if ((srch += (desty - clipy)) <= 0)
				return (0);
			srcy -= (desty - clipy);
			desty = clipy;
			}
		if (destx + srcw > (clipx + clipw))
			if ((srcw = (clipx + clipw) - destx) <= 0)
				return (0);
		if (desty + srch > (clipy + cliph))
			if ((srch = (clipy + cliph) - desty) <= 0)
				return (0);
		return (1);
		}
	void free_image();		// Free image.
	Image_window(Display *dpy, int screen_num, 
			unsigned int w, unsigned int h, 
			unsigned long bdr_color, unsigned long back_color);
public:
					// Create with proper depth.
	static Image_window *create(Display *dpy, int screen_num, 
			unsigned int w, unsigned int h, 
			unsigned long bdr_color, unsigned long back_color);
	~Image_window();
	Window get_win()		// Get handle.
		{ return win; }
	GC get_gc()
		{ return gc; }
	unsigned int get_width()	// Get dims.
		{ return width; }
	unsigned int get_height()
		{ return height; }
	int ready()			// Ready to draw?
		{ return (bits != 0); }
					// Set title.
	void set_title(char *title)
		{
		XStoreName(display, win, title);
		}
					// Indicate events we want.
	void select_input(unsigned int mask)
		{
		XSelectInput(display, win, mask);
		}
	void map()			// Display it.
		{
		XMapWindow(display, win);
		}
					// Resize event occurred.
	void resized(Display *dpy, int screen_num,
				unsigned int neww, unsigned int newh);
	void show();			// Repaint entire window.
	void clear_clip()		// Reset clip to whole window.
		{ clipx = clipy = 0; clipw = width; cliph = height; }
					// Set clip.
	void set_clip(int x, int y, int w, int h)
		{
		clipx = x;
		clipy = y;
		clipw = w;
		cliph = h;
		}
	/*
	 *	16-bit color methods.
	 */
	void exit16();			// Print error and exit.
					// Fill with given pixel.
	virtual void fill16(unsigned short pix)
		{ exit16(); }
					// Fill rect. wth pixel.
	virtual void fill16(unsigned short pix, int srcw, int srch,
						int destx, int desty)
		{ exit16(); }
					// Copy rectangle into here.
	virtual void copy16(unsigned short *src_pixels, int srcw, int srch,
						int destx, int desty)
		{ exit16(); }
					// Copy rect. with transp. color.
	virtual void copy_transparent16(unsigned char *src_pixels, int srcw,
					int srch, int destx, int desty)
		{ exit16(); }
	/*
	 *	8-bit color methods:
	 */
					// Set palette.
	virtual void set_palette(unsigned char *rgbs, int maxval, 
						int brightness = 100) = 0;
					// Fill with given (8-bit) value.
	virtual void fill8(unsigned char val) = 0;
					// Fill rect. wth pixel.
	virtual void fill8(unsigned char val, int srcw, int srch,
						int destx, int desty) = 0;
					// Copy rectangle into here.
	virtual void copy8(unsigned char *src_pixels, int srcw, int srch,
						int destx, int desty) = 0;
					// Copy rect. with transp. color.
	virtual void copy_transparent8(unsigned char *src_pixels, int srcw,
					int srch, int destx, int desty) = 0;
	/*
	 *	Depth-independent methods:
	 */
					// Copy within itself.
	virtual void copy(int srcx, int srcy, int srcw, int srch, 
						int destx, int desty) = 0;
					// Open a (truetype) font file.
	Font_face *open_font(char *fname, int points);
	void close_font(Font_face *font);// Close font.
					// Draw text in given rectangle.
	void draw_text_box(Font_face *font, char *text,
						int x, int y, int w, int h);
					// Draw text at given point.
	int draw_text(Font_face *font, char *text, int x, int y);
	int draw_text(Font_face *font, char *text, int textlen, int x, int y);
					// Draw a char. at given point.
	int draw_char(Font_face *font, short chr, int x, int y);
					// Get width of text.
	int get_text_width(Font_face *font, char *text);
	int get_text_width(Font_face *font, char *text, int textlen);
					// Get height.
	int get_text_height(Font_face *font);
	};

/*
 *	Image window on a 16-bit display:
 */
class Image_window16 : public Image_window
	{
	unsigned short *palette;	// Palette for 8-bit graphics.
	void create_default_palette();
	unsigned short *get_pixels()	// Cast bits.
		{ return (unsigned short *) bits; }
public:
	Image_window16(Display *dpy, int screen_num, 
			unsigned int w, unsigned int h, 
			unsigned long bdr_color, unsigned long back_color)
		: Image_window(dpy, screen_num, w, h, bdr_color, back_color),
			palette(0)
		{
		create_default_palette();
		}
	~Image_window16()
		{
		delete palette;
		}
					// Create pixel from r,g,b values.
	unsigned short rgb(unsigned short r, unsigned short g, 
							unsigned short b)
		{ return ((r&0x1f) << 11) + ((g&0x1f) << 6) + (b&0x1f); }
					// Set a palette color.
	void set_palette_color(unsigned char num,
		unsigned short r, unsigned short g, unsigned short b)
		{
		palette[num] = rgb(r, g, b);
		}
	/*
	 *	16-bit color methods.
	 */
					// Fill with given pixel.
	virtual void fill16(unsigned short pix);
					// Fill rect. wth pixel.
	virtual void fill16(unsigned short pix, int srcw, int srch,
						int destx, int desty);
					// Copy rectangle into here.
	virtual void copy16(unsigned short *src_pixels, int srcw, int srch,
						int destx, int desty);
	/*
	 *	Depth-independent methods:
	 */
					// Copy within itself.
	virtual void copy(int srcx, int srcy, int srcw, int srch, 
							int destx, int desty);
	/*
	 *	8-bit color methods:
	 */
					// Set palette.
	virtual void set_palette(unsigned char *rgbs, int maxval, 
						int brightness = 100);
					// Fill with given (8-bit) value.
	virtual void fill8(unsigned char val)
		{ Image_window16::fill16(palette[val]); }
					// Fill rect. wth pixel.
	virtual void fill8(unsigned char val, int srcw, int srch,
						int destx, int desty)
		{ Image_window16::fill16(
				palette[val], srcw, srch, destx, desty); }
					// Copy rectangle into here.
	virtual void copy8(unsigned char *src_pixels, int srcw, int srch,
						int destx, int desty);
					// Copy rect. with transp. color.
	virtual void copy_transparent8(unsigned char *src_pixels, int srcw,
					int srch, int destx, int desty);
	};

/*
 *	Image window on an 8-bit display:
 */
class Image_window8 : public Image_window
	{
	Colormap colormap;		// The palette.
public:
	Image_window8(Display *dpy, int screen_num, 
			unsigned int w, unsigned int h, 
			unsigned long bdr_color, unsigned long back_color)
		: Image_window(dpy, screen_num, w, h, bdr_color, back_color),
		  colormap(0)
		{  }
	~Image_window8()
		{
		if (colormap)
			XFreeColormap(display, colormap);
		}
	/*
	 *	Depth-independent methods:
	 */
					// Copy within itself.
	virtual void copy(int srcx, int srcy, int srcw, int srch, 
							int destx, int desty);
	/*
	 *	8-bit color methods:
	 */
					// Set palette.
	virtual void set_palette(unsigned char *rgbs, int maxval, 
						int brightness = 100);
					// Fill with given (8-bit) value.
	virtual void fill8(unsigned char val);
					// Fill rect. wth pixel.
	virtual void fill8(unsigned char val, int srcw, int srch,
						int destx, int desty);
					// Copy rectangle into here.
	virtual void copy8(unsigned char *src_pixels, int srcw, int srch,
						int destx, int desty);
					// Copy rect. with transp. color.
	virtual void copy_transparent8(unsigned char *src_pixels, int srcw,
					int srch, int destx, int desty);
	};

#endif	/* INCL_IMAGEWIN	*/

